/**
 * @typedef { import("./types").InstantiatedEffect } InstantiatedEffect
 * @typedef { import("./types").StrugglableEffect } StrugglableEffect
 * @typedef { import("./types").EffectMetadata } EffectMetadata
 */

/** @type {Array<StrugglableEffect>} */
export const struggle_data = [
  { effect: 'suspended', text: 'Dropped down!', difficulty: 50, add_effect_after: 'downed' },
  { effect: 'downed', text: 'Stood up!', difficulty: 5 },
  { effect: 'tied_hands', text: 'Hands freed!', difficulty: 20 },
  { effect: 'tied_legs', text: 'Legs freed!', difficulty: 10 },
  { effect: 'blindfolded', text: 'Blindfold removed!', difficulty: 4 },
  { effect: 'gagged', text: 'Gag removed!', difficulty: 4 },
  { effect: 'leashed', text: 'Collar removed!', difficulty: 10 },
  { effect: 'teased', text: 'Vibrating egg removed!', difficulty: 4 },
];

/** @type {{ [type: string]: EffectMetadata }} */
const effect_metadata = {
  downed: {
    name: 'Thrown down',
    description: "You're on the ground, unable to move until you get back up.",
    faceset_data: {
      type: 'surprised',
    },
  },
  gagged: {
    name: 'Gagged',
    description: 'Your mouth is stuffed full, stopping you from talking, eating and drinking.',
    faceset_data: {
      type: 'composed',
      fragment: 'gag',
      index: 0,
      animated: false,
    },
  },
  blindfolded: {
    name: 'Blindfolded',
    description: 'Your eyes are covered, stopping you from seeing almost anything.',
    faceset_data: {
      type: 'composed',
      fragment: 'blind',
      index: 1,
      animated: false,
    },
  },
  tied_hands: {
    name: 'Tied hands',
    description: 'Your hands are tied behind your back, stopping you from... Well, doing almost anything, really.',
    faceset_data: {
      type: 'composed',
      fragment: 'tie',
      index: 2,
      animated: true,
    },
  },
  tied_legs: {
    name: 'Tied legs',
    description: 'Your legs are tied together, forcing you to either slowly tiptoe or hop around.',
    faceset_data: {
      type: 'surprised',
    },
  },
  suspended: {
    name: 'Suspended',
    description: 'The ropes around your body are tied to the ceiling, lifting you off the ground.',
    faceset_data: {
      type: 'override',
      fragment: 'suspension',
      animated: true,
    },
  },
  teased: {
    name: 'Teased',
    description: 'A small vibrator fills your insides, periodically vibrating... Or not.',
  },
  distracted: {
    name: 'Distracted',
    description: "The stimulation you're feeling is making it very hard to focus.",
    faceset_data: {
      type: 'surprised',
    },
  },
  silenced: {
    name: 'Silenced',
    description: "You can't produce any sound from your mouth, and cannot cast any spells",
  },
  blinded: {
    name: 'Blinded',
    description: "You can't see anything out of your eyes",
    faceset_data: {
      type: 'surprised',
    },
  },
  leashed: {
    name: 'Collared',
    description:
      'A collar binds your neck, the leash dangling from it providing an easy way for others to control you.',
  },
  no_consumables: {
    name: 'No consumables',
    description: "Try as you might, you can't eat or drink right now.",
  },
  hanging: {
    name: 'Hanging',
    description: 'The rope around your neck makes it difficult to breathe',
    faceset_data: {
      type: 'surprised',
    },
  },
};

/**
 * Get the effect data from the effect type
 * @param {string} type The effect type to retrieve
 * @returns {EffectMetadata?} The full effect's data, or undefined if this effect doesn't exist.
 */
export function getEffectMetadataFromType(type) {
  return effect_metadata[type];
}

/**
 * @param {InstantiatedEffect[]} effect_list
 * @returns {boolean}
 */
export function hasMovementStoppingEffect(effect_list) {
  return effect_list.some((e) => e.type === 'downed' || e.type === 'suspended');
}

/**
 * @param {InstantiatedEffect[]} effect_list
 * @returns {boolean}
 */
export function hasSlowingEffect(effect_list) {
  return effect_list.some((e) => e.type === 'tied_legs');
}

/**
 * @param {InstantiatedEffect[]} effect_list
 * @returns {boolean}
 */
export function hasBlindingEffect(effect_list) {
  return effect_list.some((e) => e.type === 'blindfolded' || e.type === 'blinded');
}

/**
 * @param {InstantiatedEffect[]} effect_list The list of effects to check
 * @returns {StrugglableEffect | undefined} The strugglable effect, or nothing if there's no strugglable effect in the list.
 */
export function getEffectToStruggleFor(effect_list) {
  const effect = struggle_data.find((se) => effect_list.some((e) => e.type === se.effect));
  return effect;
}

/**
 *
 * @param {InstantiatedEffect[]} effect_list The list of active effects
 * @param {number} struggle_counter The current struggle counter
 * @returns {string} The result string
 */
export function getFacesetModifier(effect_list, struggle_counter) {
  let is_animated = false;
  let has_surpised = false;
  let faceset_override = '';
  let faceset_compositions = [];
  for (const active_effect of effect_list) {
    const metadata = getEffectMetadataFromType(active_effect.type);
    if (!metadata || !metadata.faceset_data) continue;
    if (metadata.faceset_data.type === 'composed') {
      is_animated ||= metadata.faceset_data.animated;
      faceset_compositions[metadata.faceset_data.index] = metadata.faceset_data.fragment;
    }
    if (metadata.faceset_data.type === 'override') {
      is_animated ||= metadata.faceset_data.animated;
      faceset_override = metadata.faceset_data.fragment;
    }
    if (metadata.faceset_data.type === 'surprised') {
      has_surpised = true;
    }
  }
  if (faceset_override) {
    if (is_animated && struggle_counter % 2) return `face_${faceset_override}_s`;
    return `face_${faceset_override}`;
  }
  if (faceset_compositions.length > 0) {
    let default_face_to_draw = 'face';
    for (const composition of faceset_compositions) {
      if (composition) {
        default_face_to_draw += '_' + composition;
      }
    }
    if (is_animated && struggle_counter % 2) default_face_to_draw += '_s';
    return default_face_to_draw;
  }
  if (has_surpised) return `face_surprised`;
  return `face_standard`;
}
