/**
 * @typedef { import("./types").LoadedMapData } LoadedMapData
 * @typedef { import("./types").Entity } Entity
 * @typedef { import("./types").ControllerSystemMethod } ControllerSystemMethod
 * @typedef { import("./types").SystemMethod } SystemMethod
 * @typedef { import("./types").Particle } Particle
 */

/**
 * Create a new game engine for the given context
 * @param {CanvasRenderingContext2D} renderContext The rendering context
 * @param {Array<Entity>} entities the entities to load
 * @param {LoadedMapData} loadedMapData The map information
 * @param {ControllerSystemMethod[]} controller_systems The methods to use for controlling the entities
 * @param {SystemMethod[]} render_systems The methods to use for rendering the entities
 */
export function Game(renderContext, entities, loadedMapData, controller_systems, render_systems) {
  const virtualCanvas = document.createElement('canvas');
  virtualCanvas.width = renderContext.canvas.width;
  virtualCanvas.height = renderContext.canvas.height;
  const context = virtualCanvas.getContext('2d');
  if (!context) throw new Error('Could not create context');
  this.showMetaText = true;
  this.fps = 0;
  this.context = context;
  this.virtualCanvas = virtualCanvas;
  this.renderContext = renderContext;
  /** @type {ControllerSystemMethod[]} */ this.controller_systems = controller_systems;
  /** @type {SystemMethod[]} */ this.renderer_systems = render_systems;
  /** @type {Array<Entity>} */ this.entities = entities;
  this.keyboard = {
    config: {
      up: ['z', 'ArrowUp'],
      down: ['s', 'ArrowDown'],
      left: ['q', 'ArrowLeft'],
      right: ['d', 'ArrowRight'],
      action: ['f', 'Spacebar', ' '],
      back: ['Escape'],
    },
    state: {
      up: false,
      down: false,
      left: false,
      right: false,
      action: false,
      back: false,
      shift: false,
    },
  };
  this.mouse = {
    x: -1,
    y: -1,
    primary_button: false,
    secondary_button: false,
  };
  /**
      @type {Array<
        | ({ type: "text", text: string } & Particle)
        | ({ type: "blink", color: string } & Particle)
      >}
     */
  this.particles = [];
  /**
     @type {Array<
      | { type: "text", textContent: string, timeout: number }
      | { type: "sound-effect", name: string, timeout: number }
      >}
     */
  this.events = [];
  const mapData = {
    offsetX: 350,
    offsetY: 350,
    render_width: 800,
    render_height: 600,
    tiles: loadedMapData,
  };
  this.mapData = mapData;
  this.renderer = {
    /**
     * @param {ImageBitmap} image
     * @param {number} x
     * @param {number} y
     * @param {number} width
     * @param {number} height
     */
    renderRawImage: function (image, x, y, width, height) {
      context.drawImage(image, x, y, width, height);
    },
    /**
     * @param {ImageBitmap} image
     * @param {number} x
     * @param {number} y
     * @param {number} width
     * @param {number} height
     */
    renderImage: function (image, x, y, width, height) {
      context.drawImage(image, x - mapData.offsetX, y - mapData.offsetY, width, height);
    },
    /**
     * @param {number} x
     * @param {number} y
     * @param {number} width
     * @param {number} height
     * @param {string} color
     */
    renderWorldRect: function (x, y, width, height, color) {
      context.fillStyle = color;
      context.strokeStyle = 'transparent';
      context.fillRect(x - mapData.offsetX, y - mapData.offsetY, width, height);
    },
    /**
     * @param {number} x
     * @param {number} y
     * @param {number} radius
     * @param {string} color
     */
    renderWorldCircle: function (x, y, radius, color) {
      context.fillStyle = color;
      context.strokeStyle = 'transparent';
      context.beginPath();
      context.ellipse(x - mapData.offsetX, y - mapData.offsetY, radius, radius, Math.PI / 4, 0, 2 * Math.PI);
      context.fill();
    },
  };
}

/**
 * Returns the position of a world point as displayed on the screen
 * @param {{ x: number, y: number }} worldPoint The world coordinates of the point to convert
 * @returns {{ x:number, y: number }} The on-screen position
 */
Game.prototype.toDisplayCoordinates = function (worldPoint) {
  return {
    x: worldPoint.x - this.mapData.offsetX,
    y: worldPoint.y - this.mapData.offsetY,
  };
};

/**
 * Converts the tiles around (x, y) from the given source ID to the target ID
 * @param {number} x The x coordinate of the search origin
 * @param {number} y The y coordinate of the search origin
 * @param {number} wx Number of pixels to search around in the X direction (from [x - wx ; x + wx])
 * @param {number} wy Number of pixels to search around in the Y direction (from [y - wy ; y + wy])
 * @param {number} id1 The ID of the source tile to convert
 * @param {number} id2 The ID to replace the source tiles with
 */
Game.prototype.convertTilesInProximity = function (x, y, wx, wy, id1, id2) {
  const firstTileX = Math.floor((x - wx - this.mapData.tiles.tile_width + 1) / this.mapData.tiles.tile_width);
  const firstTileY = Math.floor((y - wy - this.mapData.tiles.tile_width + 1) / this.mapData.tiles.tile_height);
  const lastTileX = Math.ceil((x + wx) / this.mapData.tiles.tile_width);
  const lastTileY = Math.ceil((y + wy) / this.mapData.tiles.tile_height);
  for (let y = firstTileY; y <= lastTileY; ++y) {
    const content_row = this.mapData.tiles.content[y];
    for (let x = firstTileX; x <= lastTileX; ++x) {
      const content_value = content_row ? content_row[x] || 0 : 0;
      if (content_value === id1) {
        content_row[x] = id2;
      }
    }
  }
};

/**
 * Show a small text particle with the given text diffusing upwards, at the given coordinates
 * @param {string} text The text to display
 * @param {number} x The starting X position for this particle
 * @param {number} y The starting Y position for this particle
 */
Game.prototype.createTextParticle = function (text, x, y) {
  this.particles.push({
    type: 'text',
    text: text,
    totalTimer: 100,
    currentTimer: 0,
    x: x,
    y: y,
    dirX: Math.random() * 0.4 - 0.2,
    dirY: -0.5,
    decayX: 0,
    decayY: 0.01,
  });
};

/**
 * Show a timed textbox with the given text
 * @param {string} text The text to display
 */
Game.prototype.createTextBox = function (text) {
  this.events.push({
    type: 'text',
    textContent: text,
    timeout: 300,
  });
};

/**
 * Show a small randomly diffusing particle at the given coordinate with a random color
 * @param {number} x The starting X position for this particle
 * @param {number} y The starting Y position for this particle
 */
Game.prototype.createBlinkParticle = function (x, y) {
  this.particles.push({
    type: 'blink',
    color: 'hsl(' + Math.floor(360 * Math.random()) + ', 50, 50)',
    totalTimer: 20,
    currentTimer: 0,
    x: x,
    y: y,
    dirX: Math.random() * 0.1 - 0.05,
    dirY: Math.random() * 0.1 - 0.05,
    decayX: Math.random() * 0.1 - 0.05,
    decayY: Math.random() * 0.1 - 0.05,
  });
};

/** @param {number} elapsedTicks */
Game.prototype.update = function (elapsedTicks) {
  for (const system of this.controller_systems) {
    system(this, elapsedTicks);
  }
};

/** @param {number} fps */
Game.prototype.render = function (fps) {
  // Update the FPS counter
  this.fps = fps;

  for (const system of this.renderer_systems) {
    system(this);
  }

  // Swap the contexts for double buffering
  this.renderContext.drawImage(this.virtualCanvas, 0, 0);
};

Game.prototype.showExitScreen = function () {
  this.context.fillStyle = 'black';
  this.context.strokeStyle = 'transparent';
  this.context.fillRect(0, 0, 800, 600);
  this.context.fillStyle = 'white';
  this.context.textAlign = 'center';
  this.context.font = '40px sans-serif';
  this.context.fillText('Stopped', 400, 320);
  // Swap the contexts for double buffering
  this.renderContext.drawImage(this.virtualCanvas, 0, 0);
};
