import { GAME_SPRITE_SIDE } from './assets.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").LoadedMapData } LoadedMapData
 * @typedef { import("../types").RawEntityData } RawEntityData
 */

const TILE_WIDTH = GAME_SPRITE_SIDE;
const TILE_HEIGHT = GAME_SPRITE_SIDE;

/**
 * Convert a possibly pseudo-coordinate pair into a normal coordinate pair
 * @param {string} x A string that might be a coordinate or a pseudo-coordinate (in tiles count)
 * @param {string} y A string that might be a coordinate or a pseudo-coordinate (in tiles count)
 * @returns {{ x: number, y: number }} The actual coordinates
 */
function to_coord(x, y) {
  let x_value = 0;
  if (x.startsWith('@')) {
    x_value = +x.substring(1) * TILE_WIDTH + TILE_WIDTH / 2;
  } else {
    x_value = +x;
  }
  let y_value = 0;
  if (y.startsWith('@')) {
    y_value = +y.substring(1) * TILE_HEIGHT + TILE_HEIGHT / 2;
  } else {
    y_value = +y;
  }
  return {
    x: x_value,
    y: y_value,
  };
}

/** @type {{ [key: string]: (line: string) => RawEntityData }} */
const entity_loaders = {
  '[player]': function (line) {
    const info = line.substring('[player] '.length).split(' ');
    const coords = to_coord(info[0], info[1]);
    return {
      type: 'player',
      x: coords.x,
      y: coords.y,
    };
  },
  '[door]': function (line) {
    const info = line.substring('[door] '.length).split(' ');
    const coords = to_coord(info[0], info[1]);
    return {
      type: 'door',
      x: coords.x,
      y: coords.y,
      sprite_closed: info[2],
      sprite_opened: info[3],
      is_opened: info[4] === 'opened',
    };
  },
  '[enemy_basic]': function (line) {
    const info = line.substring('[enemy_basic] '.length).split(' ');
    const coords = to_coord(info[1], info[2]);
    const j_coords = to_coord(info[3], info[4]);
    return {
      type: 'enemy_basic',
      sprite: info[0],
      x: coords.x,
      y: coords.y,
      jx: j_coords.x,
      jy: j_coords.y,
    };
  },
  '[enemy_patrol]': function (line) {
    const info = line.substring('[enemy_patrol] '.length).split(' ');
    const coords = to_coord(info[1], info[2]);
    let patrol_points_list = info.slice(3);
    /** @type {Array<{ x: number, y: number }>} */
    const patrol_points = [];
    while (patrol_points_list.length >= 2) {
      const patrol_coords = to_coord(patrol_points_list[0], patrol_points_list[1]);
      patrol_points.push(patrol_coords);
      patrol_points_list = patrol_points_list.slice(2);
    }
    return {
      type: 'enemy_patrol',
      sprite: info[0],
      x: coords.x,
      y: coords.y,
      patrol: patrol_points,
    };
  },
  '[enemy_mimic]': function (line) {
    const info = line.substring('[enemy_mimic] '.length).split(' ');
    const coords = to_coord(info[1], info[2]);
    return {
      type: 'enemy_mimic',
      sprite: info[0],
      x: coords.x,
      y: coords.y,
    };
  },
  '[gold]': function (line) {
    const info = line.substring('[gold] '.length).split(' ');
    const coords = to_coord(info[0], info[1]);
    return {
      type: 'gold',
      x: coords.x,
      y: coords.y,
      value: +info[2],
    };
  },
  '[stamina]': function (line) {
    const info = line.substring('[stamina] '.length).split(' ');
    const coords = to_coord(info[0], info[1]);
    return {
      type: 'stamina',
      x: coords.x,
      y: coords.y,
      value: +info[2],
    };
  },
  '[effect_item]': function (line) {
    const info = line.substring('[effect_item] '.length).split(' ');
    const coords = to_coord(info[1], info[2]);
    return {
      type: 'effect_item',
      sprite: info[0],
      x: coords.x,
      y: coords.y,
      duration_s: +info[3],
      is_infinite: info[3] === 'infinite',
      effect_type: info.slice(4),
    };
  },
};

/**
 * Load a map data into the given thing.
 * @param {string} raw_data the map data to load, in raw text format
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 * @returns {{
 *    tiles: LoadedMapData,
 *    entities: RawEntityData[],
 * }}
 */
export function loadRawMap(raw_data, images) {
  /** @type {{ [key: string]: number }} */
  const tile_mapping = {};
  /** @type { { sprites: string[] }[] } */
  const tile_infos = [];
  /** @type { number[][] } */
  const map_content = [];
  /** @type { RawEntityData[] } */
  const entities = [];

  /** @type {{ [key: number]: boolean }} */
  const tiles_prevent_movement = {};
  /** @type {{ [key: number]: boolean }} */
  const tiles_prevent_vision = {};

  // Simplified parser for the data.
  let is_association_data = false;
  let is_map_data = false;
  for (const line of raw_data.split('\n')) {
    // Parser for the symbol definitions
    if (line === '[/symbols]') {
      is_association_data = false;
      continue;
    }
    if (is_association_data) {
      const symbol = line[0];
      tile_mapping[symbol] = tile_infos.length;
      const movement_fragment = line.substring(2, 3);
      tiles_prevent_movement[tile_mapping[symbol]] = movement_fragment === 'b';
      const vision_fragment = line.substring(3, 4);
      tiles_prevent_vision[tile_mapping[symbol]] = vision_fragment === 'b';

      const sprites_text = line.substring(5);
      tile_infos.push({ sprites: sprites_text.split(' ') });
      continue;
    }
    if (line === '[symbols]') {
      is_association_data = true;
      continue;
    }

    // Parser for the map definitions
    if (line === '[/map]') {
      is_map_data = false;
      continue;
    }
    if (is_map_data) {
      /** @type { number[] } */
      const map_line_content = [];
      for (const entry of line) {
        const tile_value = tile_mapping[entry];
        if (tile_value === undefined) {
          console.warn(`Unknown symbol ${entry} in map data`);
        }
        map_line_content.push(tile_value || 0);
      }
      map_content.push(map_line_content);
      continue;
    }
    if (line === '[map]') {
      is_map_data = true;
      continue;
    }

    // Parser for the entity data
    for (const start_value in entity_loaders) {
      if (line.startsWith(start_value)) {
        entities.push(entity_loaders[start_value](line));
        continue;
      }
    }
  }

  /** @type {{ [key: number]: (game: Game, x: number, y: number) => void }} */
  const map_functions = {};
  tile_infos.forEach((info, index) => {
    map_functions[index] = function (game, x, y) {
      for (const sprite of info.sprites) {
        game.renderer.renderRawImage(images[sprite], x, y, TILE_WIDTH, TILE_HEIGHT);
      }
    };
  });

  return {
    tiles: {
      tile_width: TILE_WIDTH,
      tile_height: TILE_HEIGHT,
      tiles_prevent_movement: tiles_prevent_movement,
      tiles_prevent_vision: tiles_prevent_vision,
      map: map_functions,
      content: map_content,
    },
    entities: entities,
  };
}
