import { Game } from './engine.mjs';

import { createDoor } from './entities/door.mjs';
import { createEnemyBasic } from './entities/enemy-basic.mjs';
import { createEnemyMimic } from './entities/enemy-mimic.mjs';
import { createEnemyPatrol } from './entities/enemy-patrol.mjs';
import { createGoldItem } from './entities/gold-item.mjs';
import { createItemThatAddsEffects } from './entities/item-that-adds-effects.mjs';
import { createPlayer } from './entities/player.mjs';
import { createStaminaPickup } from './entities/stamina-pickup.mjs';

import { imageDescriptions } from './loader/assets.mjs';
import { loadImageObject } from './loader/resources.mjs';
import { loadRawMap } from './loader/map.mjs';

import { getBehaviorAIBasicSystem } from './system/behavior-ai-basic-system.mjs';
import { getBehaviorAIMimicSystem } from './system/behavior-ai-mimic-system.mjs';
import { getBehaviorAIComplexSystem } from './system/behavior-ai-complex-system.mjs';
import { getBehaviorKeyboardSystem } from './system/behavior-keyboard-system.mjs';
import { getControllerCameraSystem } from './system/controller-camera-system.mjs';
import { getControllerContactSystem } from './system/controller-contact-system.mjs';
import { getControllerDoorCollisionSystem } from './system/controller-door-collision-system.mjs';
import { getControllerEffectsSystem } from './system/controller-effects-system.mjs';
import { getControllerMovementSystem } from './system/controller-movement-system.mjs';
import { getControllerStaminaRegenSystem } from './system/controller-stamina-regen-system.mjs';
import { getHudEffectsSystem } from './system/hud-effects-system.mjs';
import { getHudFacesetSystem } from './system/hud-faceset-system.mjs';
import { getHudFpsSystem } from './system/hud-fps-system.mjs';
import { getHudInventorySystem } from './system/hud-inventory-system.mjs';
import { getHudPlayerStatusSystem } from './system/hud-player-status-system.mjs';
import { getHudStruggleBarSystem } from './system/hud-struggle-bar-system.mjs';
import { getHudTextEventsSystem } from './system/hud-text-events-system.mjs';
import { getRenderAngleSystem } from './system/render-angle-system.mjs';
import { getRenderBlindfoldSystem } from './system/render-blindfold-system.mjs';
import { getRenderMapSystem } from './system/render-map-system.mjs';
import { getRenderParticlesSystem } from './system/render-particles-system.mjs';
import { getRenderSimpleSpriteSystem } from './system/render-simple-sprite-system.mjs';
import { getRenderSimpleSquareSystem } from './system/render-simple-square-system.mjs';
import { getRenderVisionSystem } from './system/render-vision-system.mjs';
import { getSpriteControllerBasicEnemySystem } from './system/sprite-controller-basic-enemy-system.mjs';
import { getSpriteControllerDoorSystem } from './system/sprite-controller-door-system.mjs';
import { getSpriteControllerMimicSystem } from './system/sprite-controller-mimic-system.mjs';
import { getSpriteControllerPlayerSystem } from './system/sprite-controller-player-system.mjs';

/** @type {HTMLCanvasElement} */
// @ts-ignore (for some reason, this next line isn't cast in the proper type automatically)
const canvas = document.getElementById('game');
const context = (() => {
  const context = canvas.getContext('2d');
  if (!context) throw new Error('Could not create 2D context');
  return context;
})();

const launchWarningMessage = document.getElementById('launch_warning');
if (launchWarningMessage) {
  launchWarningMessage.remove();
}

const MS_BETWEEN_TICKS = 1000 / 20;
const MAX_MS_BETWEEN_FRAMES = 1000 / 60;

let total_image_count = 0;
let loaded_image_count = 0;

/**
 * @param {number} count The total count of images
 */
function onImageCountReady(count) {
  total_image_count = count;
  loaded_image_count = 0;
  displayImageLoadingBar();
}

function onSingleImageLoaded() {
  loaded_image_count++;
  displayImageLoadingBar();
}

function displayImageLoadingBar() {
  context.fillStyle = 'black';
  context.fillRect(0, 0, 800, 600);
  context.fillStyle = 'grey';
  context.fillRect(198, 498, 400, 24);
  context.fillStyle = 'white';
  context.fillRect(200, 500, 400, 20);
  context.font = '50px sans-serif';
  context.textBaseline = 'middle';
  context.fillText(`Loading images... ${loaded_image_count} / ${total_image_count}`, 100, 400, 600);
  context.fillStyle = 'purple';
  context.fillRect(200, 500, (400 * loaded_image_count) / total_image_count, 20);
}

Promise.all([
  fetch('assets/maps/map1.txt').then((m) => m.text()),
  loadImageObject(imageDescriptions, onImageCountReady, onSingleImageLoaded),
]).then((results) => {
  total_image_count = 0;
  const raw_map_data = results[0];
  const images = results[1];
  const loaded_map_data = loadRawMap(raw_map_data, images);

  const entities = loaded_map_data.entities.map((e) => {
    switch (e.type) {
      case 'player':
        return createPlayer(e.x, e.y);
      case 'door':
        return createDoor(e.x, e.y, e.sprite_closed, e.sprite_opened, e.is_opened);
      case 'enemy_basic':
        return createEnemyBasic(e.sprite, e.x, e.y, e.jx, e.jy);
      case 'enemy_mimic':
        return createEnemyMimic(e.sprite, e.x, e.y);
      case 'enemy_patrol':
        return createEnemyPatrol(e.sprite, e.x, e.y, e.patrol);
      case 'gold':
        return createGoldItem(e.x, e.y, e.value);
      case 'stamina':
        return createStaminaPickup(e.x, e.y, e.value);
      case 'effect_item':
        return createItemThatAddsEffects(e.x, e.y, e.sprite, e.effect_type, e.duration_s, e.is_infinite);
    }
    /** @param {never} e */
    function assertAllLoaderTypesHandled(e) {
      throw new Error(`Could not parse entity type ${e}`);
    }
    assertAllLoaderTypesHandled(e);
  });

  const controller_systems = [
    getControllerDoorCollisionSystem(),
    getControllerEffectsSystem(),
    getControllerStaminaRegenSystem(),
    getBehaviorKeyboardSystem(),
    getControllerContactSystem(),
    getBehaviorAIBasicSystem(),
    getBehaviorAIMimicSystem(),
    getBehaviorAIComplexSystem(),
    getControllerMovementSystem(),
    getSpriteControllerPlayerSystem(),
    getSpriteControllerBasicEnemySystem(),
    getSpriteControllerMimicSystem(),
    getSpriteControllerDoorSystem(),
    getControllerCameraSystem(),
  ];

  const render_systems = [
    getRenderMapSystem(),
    getRenderVisionSystem(),
    getRenderSimpleSquareSystem(),
    getRenderSimpleSpriteSystem(images),
    getRenderAngleSystem(),
    getRenderParticlesSystem(),
    getRenderBlindfoldSystem(),
    getHudFpsSystem(),
    getHudStruggleBarSystem(images),
    getHudFacesetSystem(images),
    getHudPlayerStatusSystem(),
    getHudInventorySystem(images),
    getHudEffectsSystem(images),
    getHudTextEventsSystem(),
  ];

  (function gameLoop() {
    const game = new Game(context, entities, loaded_map_data.tiles, controller_systems, render_systems);
    window.addEventListener('keydown', function (event) {
      game.keyboard.state.shift = event.shiftKey;
      for (const keyName in game.keyboard.config) {
        // Typescript doesn't seem to understand (maybe with reason?) that "keyName" is not a generic string.
        //@ts-ignore
        if (game.keyboard.config[keyName] && game.keyboard.config[keyName].indexOf(event.key) >= 0) {
          //@ts-ignore
          game.keyboard.state[keyName] = true;
        }
      }
    });

    window.addEventListener('keyup', function (event) {
      game.keyboard.state.shift = event.shiftKey;
      for (const keyName in game.keyboard.config) {
        //@ts-ignore
        if (game.keyboard.config[keyName] && game.keyboard.config[keyName].indexOf(event.key) >= 0) {
          //@ts-ignore
          game.keyboard.state[keyName] = false;
        }
      }
    });

    canvas.addEventListener('mousedown', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
      if (event.button === 0) game.mouse.primary_button = true;
      if (event.button === 2) game.mouse.secondary_button = true;
    });

    canvas.addEventListener('mousemove', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
    });

    canvas.addEventListener('mouseup', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
      if (event.button === 0) game.mouse.primary_button = false;
      if (event.button === 2) game.mouse.secondary_button = false;
    });

    let lastLoop = Date.now();
    let lastTick = Date.now();
    let shouldQuit = false;
    function singleLoop() {
      const currentLoop = Date.now();
      if (currentLoop > lastTick + MS_BETWEEN_TICKS) {
        const elapsedMsBetweenNowAndThen = currentLoop - lastTick;
        const remainingMs = elapsedMsBetweenNowAndThen % MS_BETWEEN_TICKS;
        const elapsedTicks = (elapsedMsBetweenNowAndThen - remainingMs) / MS_BETWEEN_TICKS;
        game.update(elapsedTicks);
        lastTick = currentLoop - remainingMs;
      }
      if (currentLoop > lastLoop + MAX_MS_BETWEEN_FRAMES) {
        const fps = 1000 / (currentLoop - lastLoop);
        game.render(fps);
        lastLoop = currentLoop;
      }
      if (game.keyboard.state.back) {
        shouldQuit = true;
      }
    }
    function insideLoop() {
      singleLoop();
      if (!shouldQuit) {
        window.requestAnimationFrame(insideLoop);
      } else {
        game.showExitScreen();
      }
    }
    window.requestAnimationFrame(insideLoop);
  })();
});
