import { getDistance } from '../utils/distance.mjs';
import { getBetween } from '../utils/between.mjs';
import { getVisionBox } from '../utils/vision-box.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").Entity } Entity
 * @typedef { import("../types").PositionComponent } PositionComponent
 * @typedef { import("../types").ControllerMovementComponent } ControllerMovementComponent
 * @typedef { import("../types").BehaviorAIBasicComponent } BehaviorAIBasicComponent
 * @typedef { import("../types").DirectionAngleComponent } DirectionAngleComponent
 * @typedef { import("../types").DirectionVisionComponent } DirectionVisionComponent
 * @typedef {{
 *   behavior_ai_basic_component: BehaviorAIBasicComponent,
 *   controller_movement_component: ControllerMovementComponent,
 *   angle_component: DirectionAngleComponent,
 *   position_component: PositionComponent,
 *   vision_component: DirectionVisionComponent,
 * }} FilteredControllers
 */

/**
 * @param {Entity} entity
 */
function entityCanBeCarried(entity) {
  if (!entity.basic_enemy_ai_detection_component) return false;
  if (!entity.basic_enemy_ai_detection_component.can_be_jailed) false;
  if (entity.effects_component) {
    if (!entity.effects_component.effects.some((e) => e.type === 'downed')) return false;
    if (!entity.effects_component.effects.some((e) => e.type === 'tied_hands')) return false;
    if (!entity.effects_component.effects.some((e) => e.type === 'tied_legs')) return false;
    if (!entity.effects_component.effects.some((e) => e.type === 'leashed')) return false;
  }
  return true;
}

/**
 *
 * @param {Game} game The current game
 * @param { FilteredControllers } controllers
 * @returns {{
 *  transitionTo: 'targeting' | 'searching' | 'idle' | 'carrying' | undefined,
 *  onTransition: (() => void) | undefined,
 *  target: { x: number, y: number } | undefined,
 *  target_entity: Entity | undefined,
 * }}
 */
function detectionInfo(game, controllers) {
  const vision_box = getVisionBox(
    game,
    controllers.position_component,
    controllers.angle_component.angle,
    controllers.vision_component.front,
    controllers.vision_component.sides,
    controllers.vision_component.back
  );

  for (const entity of game.entities) {
    if (!entity.basic_enemy_ai_detection_component) continue;
    if (!entity.position_component) continue;

    const can_see_target =
      entity.basic_enemy_ai_detection_component.can_be_detected &&
      vision_box.x < entity.position_component.x &&
      vision_box.x + vision_box.width > entity.position_component.x &&
      vision_box.y < entity.position_component.y &&
      vision_box.y + vision_box.height > entity.position_component.y;
    const player_can_be_carried = entity.basic_enemy_ai_detection_component.can_be_jailed && entityCanBeCarried(entity);
    const is_targeting_home =
      controllers.behavior_ai_basic_component.target.x === controllers.behavior_ai_basic_component.home.x &&
      controllers.behavior_ai_basic_component.target.y === controllers.behavior_ai_basic_component.home.y;
    const distance_to_home = getDistance(controllers.position_component, controllers.behavior_ai_basic_component.home);

    if (can_see_target && player_can_be_carried) {
      const player_distance_to_jail = getDistance(
        entity.position_component,
        controllers.behavior_ai_basic_component.jail
      );
      if (player_distance_to_jail > controllers.behavior_ai_basic_component.home_range) {
        if (controllers.behavior_ai_basic_component.state !== 'carrying') {
          // Start carrying
          return {
            transitionTo: 'carrying',
            onTransition: function () {
              game.createTextParticle('•••', controllers.position_component.x, controllers.position_component.y);
              game.createTextBox('Finally.');
            },
            target: {
              x: controllers.behavior_ai_basic_component.jail.x,
              y: controllers.behavior_ai_basic_component.jail.y,
            },
            target_entity: entity,
          };
        }

        // Keep carrying
        return {
          transitionTo: undefined,
          onTransition: undefined,
          target: undefined,
          target_entity: entity,
        };
      }
      if (
        player_distance_to_jail <= controllers.behavior_ai_basic_component.home_range &&
        distance_to_home > controllers.behavior_ai_basic_component.enemy_range
      ) {
        if (controllers.behavior_ai_basic_component.state !== 'searching') {
          return {
            transitionTo: 'searching',
            onTransition: function () {
              game.createTextParticle('•••', controllers.position_component.x, controllers.position_component.y);
              game.createTextBox('And stay there.');
              game.convertTilesInProximity(
                controllers.position_component.x,
                controllers.position_component.y,
                15,
                15,
                3,
                2
              );
            },
            target: {
              x: controllers.behavior_ai_basic_component.home.x,
              y: controllers.behavior_ai_basic_component.home.y,
            },
            target_entity: entity,
          };
        }
        return {
          transitionTo: undefined,
          onTransition: undefined,
          target: {
            x: controllers.behavior_ai_basic_component.home.x,
            y: controllers.behavior_ai_basic_component.home.y,
          },
          target_entity: entity,
        };
      }
      if (
        distance_to_home <= controllers.behavior_ai_basic_component.enemy_range &&
        controllers.behavior_ai_basic_component.state !== 'idle'
      ) {
        return {
          transitionTo: 'idle',
          onTransition: function () {
            game.createTextParticle('•••', controllers.position_component.x, controllers.position_component.y);
          },
          target: undefined,
          target_entity: entity,
        };
      }
      return {
        transitionTo: undefined,
        onTransition: undefined,
        target: undefined,
        target_entity: entity,
      };
    }

    if (can_see_target) {
      controllers.behavior_ai_basic_component.search_cooldown = 30;
      const was_idle = controllers.behavior_ai_basic_component.state === 'idle';
      const was_searching = controllers.behavior_ai_basic_component.state === 'searching';
      const was_carrying = controllers.behavior_ai_basic_component.state === 'carrying';
      if (was_idle || was_searching || was_carrying) {
        const text = was_carrying ? 'Hey! Get back!' : was_idle ? 'Hey! You there!' : 'Found you!';
        // Leave the other states and transition to the targeting state
        return {
          transitionTo: 'targeting',
          onTransition: function () {
            game.createTextParticle('!!!', controllers.position_component.x, controllers.position_component.y);
            game.createTextBox(text);
          },
          target: { x: entity.position_component.x, y: entity.position_component.y },
          target_entity: entity,
        };
      }

      // Target the player position
      return {
        transitionTo: undefined,
        onTransition: undefined,
        target: { x: entity.position_component.x, y: entity.position_component.y },
        target_entity: entity,
      };
    }

    // The enemy cannot see the player
    if (controllers.behavior_ai_basic_component.state === 'carrying') {
      // Enter search
      return {
        transitionTo: 'searching',
        onTransition: function () {
          game.createTextBox('Where did they go?!');
          game.createTextParticle('?', controllers.position_component.x, controllers.position_component.y);
        },
        target: undefined,
        target_entity: entity,
      };
    }

    if (controllers.behavior_ai_basic_component.state === 'targeting') {
      controllers.behavior_ai_basic_component.search_cooldown--;
      if (controllers.behavior_ai_basic_component.search_cooldown <= 0) {
        // Enter search
        return {
          transitionTo: 'searching',
          onTransition: function () {
            game.createTextBox('Where did they go?!');
            game.createTextParticle('?', controllers.position_component.x, controllers.position_component.y);
          },
          target: undefined,
          target_entity: entity,
        };
      }
      return {
        transitionTo: undefined,
        onTransition: undefined,
        target: undefined,
        target_entity: entity,
      };
    }

    if (!is_targeting_home && distance_to_home > controllers.behavior_ai_basic_component.enemy_range) {
      const distance_to_target_pos = getDistance(
        controllers.position_component,
        controllers.behavior_ai_basic_component.target
      );
      if (controllers.behavior_ai_basic_component.state !== 'searching') {
        if (distance_to_target_pos <= controllers.behavior_ai_basic_component.enemy_range) {
          // Start searching towards home
          return {
            transitionTo: 'searching',
            onTransition: function () {},
            target: {
              x: controllers.behavior_ai_basic_component.home.x,
              y: controllers.behavior_ai_basic_component.home.y,
            },
            target_entity: entity,
          };
        }
        // Start searching towards home
        return {
          transitionTo: 'searching',
          onTransition: function () {},
          target: undefined,
          target_entity: entity,
        };
      }
      if (distance_to_target_pos <= controllers.behavior_ai_basic_component.enemy_range) {
        // Keep searching but re-target home
        return {
          transitionTo: undefined,
          onTransition: undefined,
          target: {
            x: controllers.behavior_ai_basic_component.home.x,
            y: controllers.behavior_ai_basic_component.home.y,
          },
          target_entity: entity,
        };
      }
      // Keep searching
      return {
        transitionTo: undefined,
        onTransition: undefined,
        target: undefined,
        target_entity: entity,
      };
    }

    if (
      controllers.behavior_ai_basic_component.state === 'searching' &&
      distance_to_home <= controllers.behavior_ai_basic_component.enemy_range
    ) {
      // Idle at home
      return {
        transitionTo: 'idle',
        onTransition: function () {
          game.createTextBox('Hmm...');
          game.createTextParticle('•••', controllers.position_component.x, controllers.position_component.y);
        },
        target: undefined,
        target_entity: undefined,
      };
    }
  }

  return {
    transitionTo: undefined,
    onTransition: undefined,
    target: undefined,
    target_entity: undefined,
  };
}

/**
 * @param {Game} game
 * @param {Array<{ effect: string, text: string }>} effects_to_add
 * @param {Entity} entity
 */
function addRandomEffectToTargetEntity(game, effects_to_add, entity) {
  if (!entity.position_component) return;
  if (!entity.effects_component) return;
  const effects_component = entity.effects_component;

  if (!effects_component.effects.some((e) => e.type === 'blinded')) {
    effects_component.effects.push({
      type: 'blinded',
      isInfinite: false,
      remainingFrameCount: 20 * 10,
    });
    game.createTextParticle('Blinded!', entity.position_component.x, entity.position_component.y);
  }

  let restraint = effects_to_add.find((r) => !effects_component.effects.some((e) => e.type === r.effect));
  if (restraint) {
    effects_component.effects.push({
      type: restraint.effect,
      isInfinite: true,
    });
    game.createTextParticle(restraint.text, entity.position_component.x, entity.position_component.y);
  }
}

const basicEnemyAIPatterns = {
  /**
   * @param { Game } game
   * @param { FilteredControllers } controllers
   * @param { number } elapsedTicks
   */
  idle: function (game, controllers, elapsedTicks) {
    const result = detectionInfo(game, controllers);
    if (result.transitionTo && result.onTransition) {
      controllers.behavior_ai_basic_component.state = result.transitionTo;
      result.onTransition();
    }

    controllers.controller_movement_component.delta_x = 0;
    controllers.controller_movement_component.delta_y = 0;

    controllers.behavior_ai_basic_component.target = result.target || controllers.behavior_ai_basic_component.target;
    controllers.behavior_ai_basic_component.cycle += elapsedTicks;
    if (controllers.behavior_ai_basic_component.cycle < 2) {
      return;
    }
    controllers.behavior_ai_basic_component.cycle -= 2;

    const modifier = Math.PI / 32;
    controllers.angle_component.angle += modifier;
    if (controllers.angle_component.angle < 0) {
      controllers.angle_component.angle += 2 * Math.PI;
    }
    controllers.angle_component.angle %= 2 * Math.PI;
  },
  /**
   * @param { Game } game
   * @param { FilteredControllers } controllers
   * @param { number } elapsedTicks
   */
  carrying: function (game, controllers, elapsedTicks) {
    const result = detectionInfo(game, controllers);
    if (result.transitionTo && result.onTransition) {
      controllers.behavior_ai_basic_component.state = result.transitionTo;
      result.onTransition();
    }

    if (!result.target_entity) return;

    // Drag the player to home
    const target_angle = Math.atan2(
      controllers.position_component.y - controllers.behavior_ai_basic_component.jail.y,
      controllers.position_component.x - controllers.behavior_ai_basic_component.jail.x
    );
    controllers.controller_movement_component.delta_x -=
      controllers.behavior_ai_basic_component.dragging_speed * Math.cos(target_angle);
    controllers.controller_movement_component.delta_y -=
      controllers.behavior_ai_basic_component.dragging_speed * Math.sin(target_angle);

    // Force-drag the player away
    // TODO: Rework the way this is done by using a specific "dragging" component that interferes with the player's movement component.
    if (result.target_entity.position_component) {
      result.target_entity.position_component.x -=
        controllers.behavior_ai_basic_component.dragging_speed * Math.cos(target_angle);
      result.target_entity.position_component.y -=
        controllers.behavior_ai_basic_component.dragging_speed * Math.sin(target_angle);
    }

    controllers.angle_component.angle = target_angle;
    if (controllers.angle_component.angle < 0) {
      controllers.angle_component.angle += 2 * Math.PI;
    }
    controllers.angle_component.angle %= 2 * Math.PI;
  },
  /**
   * @param { Game } game
   * @param { FilteredControllers } controllers
   * @param { number } elapsedTicks
   */
  targeting: function (game, controllers, elapsedTicks) {
    const result = detectionInfo(game, controllers);
    if (result.transitionTo && result.onTransition) {
      controllers.behavior_ai_basic_component.state = result.transitionTo;
      result.onTransition();
    }

    if (!result.target_entity) return;

    const player_in_attack_range = result.target_entity.position_component
      ? getDistance(controllers.position_component, result.target_entity.position_component) <=
        controllers.behavior_ai_basic_component.enemy_range
      : false;
    const can_attack = controllers.behavior_ai_basic_component.attack_cooldown == 0;
    if (result.target_entity.position_component && player_in_attack_range && can_attack) {
      // Attack the player
      addRandomEffectToTargetEntity(game, controllers.behavior_ai_basic_component.effects_to_add, result.target_entity);
      const contact = getBetween(controllers.position_component, result.target_entity.position_component);
      game.createBlinkParticle(contact.x, contact.y);
      game.createBlinkParticle(contact.x, contact.y);
      game.createBlinkParticle(contact.x, contact.y);
      controllers.behavior_ai_basic_component.attack_cooldown = 10;
    }

    controllers.behavior_ai_basic_component.target = result.target || controllers.behavior_ai_basic_component.target;
    controllers.behavior_ai_basic_component.cycle += elapsedTicks;
    if (controllers.behavior_ai_basic_component.cycle < 2) {
      return;
    }
    controllers.behavior_ai_basic_component.cycle -= 2;
    const distance_to_target_pos = getDistance(
      controllers.position_component,
      controllers.behavior_ai_basic_component.target
    );
    if (distance_to_target_pos > controllers.behavior_ai_basic_component.enemy_range) {
      // Move directly towards the target position
      const target_angle = Math.atan2(
        controllers.position_component.y - controllers.behavior_ai_basic_component.target.y,
        controllers.position_component.x - controllers.behavior_ai_basic_component.target.x
      );
      controllers.controller_movement_component.delta_x =
        -controllers.behavior_ai_basic_component.speed * Math.cos(target_angle);
      controllers.controller_movement_component.delta_y =
        -controllers.behavior_ai_basic_component.speed * Math.sin(target_angle);
      controllers.angle_component.angle = target_angle;
      if (controllers.angle_component.angle < 0) {
        controllers.angle_component.angle += 2 * Math.PI;
      }
      controllers.angle_component.angle %= 2 * Math.PI;
    }
  },
  /**
   * @param { Game } game
   * @param { FilteredControllers } controllers
   * @param { number } elapsedTicks
   */
  searching: function (game, controllers, elapsedTicks) {
    const result = detectionInfo(game, controllers);
    if (result.transitionTo && result.onTransition) {
      controllers.behavior_ai_basic_component.state = result.transitionTo;
      result.onTransition();
    }
    controllers.behavior_ai_basic_component.target = result.target || controllers.behavior_ai_basic_component.target;
    controllers.behavior_ai_basic_component.cycle += elapsedTicks;
    if (controllers.behavior_ai_basic_component.cycle < 2) {
      return;
    }
    controllers.behavior_ai_basic_component.cycle -= 2;
    const distance_to_target_pos = getDistance(
      controllers.position_component,
      controllers.behavior_ai_basic_component.target
    );
    if (distance_to_target_pos > controllers.behavior_ai_basic_component.enemy_range) {
      // Move directly towards the target position
      const target_angle = Math.atan2(
        controllers.position_component.y - controllers.behavior_ai_basic_component.target.y,
        controllers.position_component.x - controllers.behavior_ai_basic_component.target.x
      );
      controllers.controller_movement_component.delta_x =
        -controllers.behavior_ai_basic_component.speed * Math.cos(target_angle);
      controllers.controller_movement_component.delta_y =
        -controllers.behavior_ai_basic_component.speed * Math.sin(target_angle);
      controllers.angle_component.angle = target_angle;
      if (controllers.angle_component.angle < 0) {
        controllers.angle_component.angle += 2 * Math.PI;
      }
      controllers.angle_component.angle %= 2 * Math.PI;
    }
  },
};

export function getBehaviorAIBasicSystem() {
  /**
   * @param {Game} game
   * @param {number} elapsedTicks
   */
  return function behaviorAIBasicSystem(game, elapsedTicks) {
    for (const entity of game.entities) {
      if (!entity.behavior_ai_basic_component) continue;
      if (!entity.controller_movement_component) continue;
      if (!entity.position_component) continue;
      if (!entity.angle_component) continue;
      if (!entity.vision_component) continue;
      if (entity.behavior_ai_basic_component.attack_cooldown > 0) {
        entity.behavior_ai_basic_component.attack_cooldown--;
      }
      basicEnemyAIPatterns[entity.behavior_ai_basic_component.state](
        game,
        {
          behavior_ai_basic_component: entity.behavior_ai_basic_component,
          controller_movement_component: entity.controller_movement_component,
          angle_component: entity.angle_component,
          position_component: entity.position_component,
          vision_component: entity.vision_component,
        },
        elapsedTicks
      );
    }
  };
}
