import { getDistance } from '../utils/distance.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").Entity } Entity
 * @typedef { import('../types').BehaviorAIMimicComponent } ControllerMimicAIComponent
 */

/**
 * @param {Entity} entity
 */
function isValidCaptureTarget(entity) {
  if (entity.contained_component && entity.contained_component.is_inside_container) return false;
  if (entity.effects_component && entity.effects_component.effects.some((e) => e.type === 'leashed')) return false;
  return true;
}

export function getBehaviorAIMimicSystem() {
  /**
   * @param {Game} game
   * @param {number} elapsedTicks
   */
  return function behaviorAIMimicSystem(game, elapsedTicks) {
    for (const entity of game.entities) {
      if (!entity.behavior_ai_mimic_component) continue;
      if (!entity.position_component) continue;

      const controller = entity.behavior_ai_mimic_component;

      controller.animation_cycle += elapsedTicks;
      controller.animation_cycle %= 8;

      /** @type { Array<{ entity: Entity, is_capturable: boolean, distance: number }> } */
      const targets = [];
      for (const mimic_target of game.entities) {
        if (!mimic_target.mimic_ai_detection_component) continue;
        if (!mimic_target.mimic_ai_detection_component.can_be_detected) continue;
        if (!mimic_target.position_component) continue;

        const targetDistance = getDistance(entity.position_component, mimic_target.position_component);
        targets.push({
          entity: mimic_target,
          distance: targetDistance,
          is_capturable: mimic_target.mimic_ai_detection_component.can_be_captured,
        });
      }

      if (!targets.length) {
        controller.state = 'idle';
        continue;
      }

      const closestValidDistance = Math.min(
        ...targets.filter((t) => isValidCaptureTarget(t.entity)).map((t) => t.distance)
      );
      if (controller.state === 'idle') {
        const target_in_active_range = closestValidDistance <= controller.active_range;
        if (target_in_active_range) {
          controller.state = 'grabbing';
        }
        continue;
      }

      if (controller.state === 'grabbing') {
        const target_in_active_range = closestValidDistance <= controller.active_range;
        if (!target_in_active_range) {
          controller.state = 'idle';
        }

        const grabbableTargets = targets.filter((t) => t.is_capturable && t.distance <= controller.grab_range);
        for (const target of grabbableTargets) {
          if (!isValidCaptureTarget(target.entity)) continue;
          if (!target.entity.position_component) continue;
          controller.digest_cycle = controller.digest_time;
          controller.state = 'eating';
          controller.last_target_position.x = target.entity.position_component.x;
          controller.last_target_position.y = target.entity.position_component.y;
          if (target.entity.contained_component) target.entity.contained_component.is_inside_container = true;
        }
        continue;
      }

      if (controller.state === 'eating') {
        controller.digest_cycle -= elapsedTicks;
        if (controller.digest_cycle <= 0) {
          controller.digest_cycle = 0;
          controller.state = 'spitting';
        }
        const grabbableTargets = targets.filter((t) => t.is_capturable && t.distance <= controller.grab_range);
        for (const target of grabbableTargets) {
          if (!target.entity.position_component) continue;
          if (target.entity.contained_component && !target.entity.contained_component.is_inside_container) continue;

          let offset = controller.digest_cycle % controller.digest_rumble === 0 ? 2 : 0;
          let angle = Math.random() * 2 * Math.PI;
          target.entity.position_component.x = entity.position_component.x + Math.cos(angle) * offset;
          target.entity.position_component.y = entity.position_component.y + Math.sin(angle) * offset;

          applyEffects(game, controller, target.entity);
        }
        continue;
      }

      if (controller.state === 'spitting') {
        const grabbableTargets = targets.filter((t) => t.is_capturable && t.distance <= controller.grab_range);
        for (const target of grabbableTargets) {
          if (!target.entity.position_component) continue;
          if (target.entity.contained_component) target.entity.contained_component.is_inside_container = false;
          target.entity.position_component.x = controller.last_target_position.x;
          target.entity.position_component.y = controller.last_target_position.y;
        }
        controller.state = 'idle';
        continue;
      }
    }
  };

  /**
   * @param {Game} game
   * @param {ControllerMimicAIComponent} controller
   * @param {Entity} target_entity
   */
  function applyEffects(game, controller, target_entity) {
    if (!target_entity.position_component) return;
    if (!target_entity.effects_component) return;
    for (const effect of controller.effects) {
      if (effect.timer < controller.digest_cycle) continue;
      if (target_entity.effects_component.effects.some((e) => e.type === effect.type)) continue;

      target_entity.effects_component.effects.push({
        type: effect.type,
        isInfinite: true,
      });
      game.createTextParticle(effect.text, target_entity.position_component.x, target_entity.position_component.y);
    }
  }
}
