import { hasMovementStoppingEffect, hasSlowingEffect } from '../effects.mjs';
import { getEffectToStruggleFor } from '../effects.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").Entity } Entity
 */

export function getBehaviorKeyboardSystem() {
  /**
   * @param {Game} game
   * @param {number} elapsedTicks
   */
  return function behaviorKeyboardSystem(game, elapsedTicks) {
    for (const entity of game.entities) {
      handleMovementSystem(game, entity, elapsedTicks);
      handleStruggleSystem(game, entity, elapsedTicks);
    }
  };

  /**
   * @param {Game} game
   * @param {Entity} entity
   * @param {number} elapsedTicks
   */
  function handleMovementSystem(game, entity, elapsedTicks) {
    if (!entity.behavior_keyboard_component) return;
    if (!entity.controller_movement_component) return;

    const can_move = (() => {
      if (entity.behavior_keyboard_component.struggle_mode === 'active') return false;
      if (entity.effects_component && hasMovementStoppingEffect(entity.effects_component.effects)) return false;
      return true;
    })();

    const can_run = (() => {
      if (!entity.stamina_component) return false;
      if (entity.stamina_component.stamina < entity.behavior_keyboard_component.run_stamina_cost) return false;
      if (entity.effects_component && hasSlowingEffect(entity.effects_component.effects)) return false;
      return true;
    })();

    const is_running = game.keyboard.state.shift && can_run;
    const base_player_speed = (() => {
      if (!can_move) return 0;
      if (!entity.behavior_keyboard_component) return 0;
      if (entity.effects_component && hasSlowingEffect(entity.effects_component.effects))
        return entity.behavior_keyboard_component.slowed_speed;
      if (is_running) return entity.behavior_keyboard_component.running_speed;
      return entity.behavior_keyboard_component.normal_speed;
    })();

    const player_speed = base_player_speed * elapsedTicks;
    let delta_x = 0;
    let delta_y = 0;
    // Move the player if needed
    if (game.keyboard.state.up && !game.keyboard.state.down && game.keyboard.state.left && !game.keyboard.state.right) {
      entity.behavior_keyboard_component.direction = 'up';
      delta_x = -player_speed / 2;
      delta_y = -player_speed / 2;
    } else if (
      game.keyboard.state.up &&
      !game.keyboard.state.down &&
      !game.keyboard.state.left &&
      game.keyboard.state.right
    ) {
      entity.behavior_keyboard_component.direction = 'up';
      delta_x = player_speed / 2;
      delta_y = -player_speed / 2;
    } else if (
      !game.keyboard.state.up &&
      game.keyboard.state.down &&
      game.keyboard.state.left &&
      !game.keyboard.state.right
    ) {
      entity.behavior_keyboard_component.direction = 'down';
      delta_x = -player_speed / 2;
      delta_y = player_speed / 2;
    } else if (
      !game.keyboard.state.up &&
      game.keyboard.state.down &&
      !game.keyboard.state.left &&
      game.keyboard.state.right
    ) {
      entity.behavior_keyboard_component.direction = 'down';
      delta_x = player_speed / 2;
      delta_y = player_speed / 2;
    } else if (game.keyboard.state.up && !game.keyboard.state.down) {
      entity.behavior_keyboard_component.direction = 'up';
      delta_y = -player_speed;
    } else if (game.keyboard.state.down && !game.keyboard.state.up) {
      entity.behavior_keyboard_component.direction = 'down';
      delta_y = player_speed;
    } else if (game.keyboard.state.left && !game.keyboard.state.right) {
      entity.behavior_keyboard_component.direction = 'left';
      delta_x = -player_speed;
    } else if (game.keyboard.state.right && !game.keyboard.state.left) {
      entity.behavior_keyboard_component.direction = 'right';
      delta_x = player_speed;
    }

    entity.controller_movement_component.delta_x = delta_x;
    entity.controller_movement_component.delta_y = delta_y;
    if (is_running)
      entity.controller_movement_component.stamina_consumption = entity.behavior_keyboard_component.run_stamina_cost;
    else entity.controller_movement_component.stamina_consumption = 0;
  }

  /**
   * @param {Game} game
   * @param {Entity} entity
   * @param {number} elapsedTicks
   */
  function handleStruggleSystem(game, entity, elapsedTicks) {
    if (!entity.behavior_keyboard_component) return;
    if (!entity.effects_component) return;
    const strugglable_effect = getEffectToStruggleFor(entity.effects_component.effects);
    if (!strugglable_effect) {
      entity.behavior_keyboard_component.struggle_counter = 0;
      entity.behavior_keyboard_component.struggle_mode = 'inactive';
      if (entity.render_struggle_bar_component) {
        entity.render_struggle_bar_component.enabled = false;
      }
      return;
    }

    if (!game.keyboard.state.action) {
      entity.behavior_keyboard_component.struggle_mode = 'inactive';
      entity.behavior_keyboard_component.previous_struggle_keys = [];
      if (entity.render_struggle_bar_component) {
        entity.render_struggle_bar_component.enabled = true;
        entity.render_struggle_bar_component.activated = false;
      }
      return;
    }

    entity.behavior_keyboard_component.struggle_mode = 'active';

    /** @type {Array<"up" | "left" | "down" | "right">} */
    const states = ['up', 'left', 'down', 'right'];
    const previousKeyStates = entity.behavior_keyboard_component.previous_struggle_keys;
    const nextKeyStates = states.filter((s) => game.keyboard.state[s]);
    const newStates = nextKeyStates.filter((s) => !previousKeyStates.includes(s));
    entity.behavior_keyboard_component.previous_struggle_keys = nextKeyStates;

    if (newStates.length && entity.stamina_component) {
      if (entity.controller_stamina_regen_component) {
        entity.controller_stamina_regen_component.stamina_regen_cooldown =
          entity.controller_stamina_regen_component.max_stamina_regen_cooldown;
      }
      entity.stamina_component.stamina -= entity.behavior_keyboard_component.struggle_stamina_cost;
      if (entity.stamina_component.stamina <= 0) {
        entity.stamina_component.stamina = 0;
      } else {
        entity.behavior_keyboard_component.struggle_counter++;
      }
    }

    if (entity.behavior_keyboard_component.struggle_counter >= strugglable_effect.difficulty) {
      entity.effects_component.effects = entity.effects_component.effects.filter(
        (e) => e.type !== strugglable_effect.effect
      );
      if (entity.position_component) {
        game.createTextParticle(strugglable_effect.text, entity.position_component.x, entity.position_component.y);
      }
      if (strugglable_effect.add_effect_after) {
        entity.effects_component.effects.push({
          type: strugglable_effect.add_effect_after,
          isInfinite: true,
        });
      }

      entity.behavior_keyboard_component.struggle_counter = 0;
    }

    if (entity.render_struggle_bar_component) {
      entity.render_struggle_bar_component.enabled = true;
      entity.render_struggle_bar_component.activated = true;
      entity.render_struggle_bar_component.progress_percent =
        entity.behavior_keyboard_component.struggle_counter / strugglable_effect.difficulty;
      entity.render_struggle_bar_component.keys = nextKeyStates;
    }
  }
}
