import { getEffectMetadataFromType } from '../effects.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").InstantiatedEffect } InstantiatedEffect
 */

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getHudEffectsSystem(images) {
  /**
   * @param {Game} game The current game
   */
  return function hudEffectsSystem(game) {
    for (const entity of game.entities) {
      if (!entity.render_hud_component) continue;
      if (!entity.render_hud_component.enable_effects) continue;
      if (!entity.effects_component) continue;
      let mouseHoveredEffect = drawActiveEffects(game, entity.effects_component.effects, 160, 450, images);
      if (mouseHoveredEffect) drawEffectPopup(game, mouseHoveredEffect);
    }
  };

  /**
   *
   * @param {Game} game
   * @param {InstantiatedEffect[]} active_effects The active effects to display
   * @param {number} startX The X start position of the active effects
   * @param {number} startY The Y start position of the active effects
   * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
   * @returns {InstantiatedEffect | null } The instantiated effect currently hovered, or `null` if there's no such effect.
   */
  function drawActiveEffects(game, active_effects, startX, startY, images) {
    let effectXOffset = startX;
    let mouseHoveredEffect = null;
    for (const effect of active_effects) {
      game.context.fillStyle = 'red';
      game.context.strokeStyle = 'black';
      game.context.fillRect(effectXOffset, startY, 40, 40);
      let effect_image = 'effect_' + effect.type;
      if (images[effect_image]) {
        game.context.drawImage(images[effect_image], effectXOffset, startY, 40, 40);
      } else {
        game.context.fillStyle = 'white';
        game.context.strokeStyle = 'transparent';
        game.context.font = '20px sans-serif';
        game.context.fillText('?', effectXOffset + 10, startY + 20);
      }
      if (!effect.isInfinite) {
        game.context.fillStyle = 'black';
        game.context.strokeStyle = 'transparent';
        game.context.font = '15px sans-serif';
        const remaining_time_s = effect.remainingFrameCount / 20;
        let text_to_draw = Math.floor(remaining_time_s) + 's';
        if (remaining_time_s < 1) {
          text_to_draw = '0.' + Math.floor(remaining_time_s * 10) + 's';
        }
        game.context.fillText(
          text_to_draw,
          effectXOffset + 40 - game.context.measureText(text_to_draw).width,
          startY + 40
        );
      }
      if (
        game.mouse.x >= effectXOffset &&
        game.mouse.x < effectXOffset + 40 &&
        game.mouse.y >= startY &&
        game.mouse.y < startY + 40
      ) {
        mouseHoveredEffect = effect;
      }
      effectXOffset += 50;
    }
    return mouseHoveredEffect;
  }

  /**
   *
   * @param {Game} game The game engine
   * @param {InstantiatedEffect} mouseHoveredEffect The effect currently hovered
   */
  function drawEffectPopup(game, mouseHoveredEffect) {
    const effect_metadata = getEffectMetadataFromType(mouseHoveredEffect.type);
    if (!effect_metadata) return;
    game.context.font = '15px sans-serif';
    const box_width = Math.max(200, 20 + game.context.measureText(effect_metadata.description).width);
    game.context.fillStyle = 'black';
    game.context.strokeStyle = 'lightgrey';
    game.context.fillRect(game.mouse.x, game.mouse.y, box_width, 70);
    game.context.strokeRect(game.mouse.x, game.mouse.y, box_width, 70);
    game.context.fillStyle = 'white';
    game.context.strokeStyle = 'transparent';
    game.context.font = '20px sans-serif';
    game.context.fillText(effect_metadata.name, game.mouse.x + 10, game.mouse.y + 20);
    game.context.font = '15px sans-serif';
    game.context.fillText(effect_metadata.description, game.mouse.x + 10, game.mouse.y + 40);
    if (mouseHoveredEffect.isInfinite) {
      game.context.fillText('Permanent effect', game.mouse.x + 10, game.mouse.y + 60);
    } else {
      game.context.fillText(
        'Remaining: ' + mouseHoveredEffect.remainingFrameCount / 20 + 's',
        game.mouse.x + 10,
        game.mouse.y + 60
      );
    }
  }
}
