import { Game } from './engine.mjs';

export type ControllerSystemMethod = (game: Game, elapsedTicks: number) => void;
export type SystemMethod = (game: Game) => void;

export type ImageAssetPartialInformation = {
  origin: { x: number; y: number; w: number; h: number };
  target: { w: number; h: number };
};

export type LoadedMapData = {
  tile_width: number;
  tile_height: number;
  map: { [key: number]: (game: Game, x: number, y: number) => void };
  tiles_prevent_movement: { [key: number]: boolean };
  tiles_prevent_vision: { [key: number]: boolean };
  /** content[y][x] */
  content: number[][];
};

export type RawEntityData =
  | { type: 'player'; x: number; y: number }
  | { type: 'door'; x: number; y: number; sprite_opened: string; sprite_closed: string; is_opened: boolean }
  | { type: 'stamina' | 'gold'; x: number; y: number; value: number }
  | {
      type: 'effect_item';
      sprite: string;
      x: number;
      y: number;
      duration_s: number;
      is_infinite: boolean;
      effect_type: string[];
    }
  | { type: 'enemy_basic'; sprite: string; x: number; y: number; jx: number; jy: number }
  | { type: 'enemy_patrol'; sprite: string; x: number; y: number; patrol: Array<{ x: number; y: number }> }
  | { type: 'enemy_mimic'; sprite: string; x: number; y: number };

export type Particle = {
  x: number;
  y: number;
  dirX: number;
  dirY: number;
  decayX: number;
  decayY: number;
  totalTimer: number;
  currentTimer: number;
};

export type EffectMetadata = {
  name: string;
  description: string;
  faceset_data?:
    | {
        type: 'surprised';
      }
    | {
        type: 'composed';
        fragment: string;
        index: number;
        animated: boolean;
      }
    | {
        type: 'override';
        fragment: string;
        animated: boolean;
      };
};

export type InstantiatedEffect = {
  type: string;
} & ({ isInfinite: true } | { isInfinite?: false; remainingFrameCount: number });

export type StrugglableEffect = {
  effect: string;
  text: string;
  difficulty: number;
  add_effect_after?: string;
};

export type AIState = 'scan' | 'search' | 'patrol' | 'capture' | 'carry' | 'back_to_home';

export type AIStatusCondition = {
  is_in_state?: AIState;
  is_not_in_state?: AIState;
  is_at_home?: boolean;
  is_carrying_target?: boolean;
  has_visible_target?: boolean;
  is_in_capture_range?: boolean;
  has_finished_scan_cycle?: boolean;
  has_reached_patrol_point?: boolean;
};

export type AIBehaviorHandler = {
  /** Used to select the next state. First matching rule gets applied */
  next_state_selector: Array<{ status: AIStatusCondition; state: AIState }>;
};

export type AICurrentStatus = {
  current_state: AIState;
  target?: {
    x: number;
    y: number;
  };
  tick_counter: number;
  increment_counter: number;
  current_patrol_index: number;
};

export type AIConfiguration = {
  scan_in_place?: {
    ticks_before_increment: number;
    rotation_percent_per_increment: number;
    increments_before_event: number;
  };
  return_to_home?: {
    x: number;
    y: number;
    range_before_reached: number;
  };
  patrol?: {
    speed: number;
    waypoints: Array<{ x: number; y: number }>;
    range_until_point_reached: number;
  };
  capture: {
    dragging_speed: number;
  };
  home?: {
    x: number;
    y: number;
  };
  jail?: {
    x: number;
    y: number;
  };
};

/* ==============================
 *          Components
 * ==============================
 */

// Behavior components

export type BehaviorAIComplexComponent = {
  behavior: AIBehaviorHandler;
  status: AICurrentStatus;
  configuration: AIConfiguration;
};

export type BehaviorAIBasicComponent = {
  state: 'idle' | 'targeting' | 'searching' | 'carrying';
  cycle: number;
  search_cooldown: number;
  attack_cooldown: number;
  speed: number;
  dragging_speed: number;
  enemy_range: number;
  home_range: number;
  home: {
    x: number;
    y: number;
  };
  jail: {
    x: number;
    y: number;
  };
  target: {
    x: number;
    y: number;
  };
  effects_to_add: Array<{
    effect: string;
    text: string;
  }>;
};

export type BehaviorAIMimicComponent = {
  state: 'idle' | 'grabbing' | 'eating' | 'spitting';
  last_target_position: {
    x: number;
    y: number;
  };
  animation_cycle: number;
  digest_cycle: number;
  digest_time: number;
  digest_rumble: number;
  active_range: number;
  grab_range: number;
  effects: Array<{
    timer: number;
    type: string;
    text: string;
  }>;
};

export type BehaviorAIPatrolComponent = {
  speed: number;
  patrol_point_range: number;
  current_patrol_index: number;
  patrol_points: Array<{ x: number; y: number }>;
};

export type BehaviorKeyboardComponent = {
  enabled: boolean;
  direction: 'up' | 'down' | 'left' | 'right';
  run_stamina_cost: number;
  slowed_speed: number;
  normal_speed: number;
  running_speed: number;
  struggle_mode: 'inactive' | 'active';
  previous_struggle_keys: Array<'up' | 'left' | 'down' | 'right'>;
  struggle_counter: number;
  struggle_stamina_cost: number;
};

// Controller components

export type ControllerCameraComponent = {
  camera_follows_entity: boolean;
};

export type ControllerDoorComponent = {
  is_opened: boolean;
  sprite_closed: string;
  sprite_opened: string;
};

export type ControllerMovementComponent = {
  enabled: boolean;
  has_collision: boolean;
  delta_x: number;
  delta_y: number;
  stamina_consumption: number;
};

export type ControllerStaminaRegenComponent = {
  stamina_regen_quantity_per_tick: number;
  stamina_regen_cooldown: number;
  max_stamina_regen_cooldown: number;
};

// Generic / Data components

export type PositionComponent = {
  x: number;
  y: number;
};

export type BasicEnemyAIDetectionComponent = {
  can_be_detected: boolean;
  can_be_jailed: boolean;
};

export type MimicAIDetectionComponent = {
  can_be_detected: boolean;
  can_be_captured: boolean;
};

export type PlayerFacesetComponent = {
  default: string;
  image: string;
};

export type StaminaComponent = {
  stamina: number;
  max_stamina: number;
};

export type ContainedComponent = {
  is_inside_container: boolean;
};

export type CollisionComponent = {
  /** Whether the collision is enabled for this component */
  is_enabled: boolean;
  /** The offset to add to the X position to start the collision */
  x_offset: number;
  /** The offset to add to the Y position to start the collision */
  y_offset: number;
  /** The collision bounding box width */
  width: number;
  /** The collision bounding box height */
  height: number;
};

export type PlayerInventoryComponent = {
  inventory: Array<{ type: 'gold'; quantity: number }>;
};

export type PlayerEffectsComponent = {
  effects: InstantiatedEffect[];
};

export type PlayerContactComponent = {
  hitbox_area: number;
};

export type DirectionVisionComponent = {
  front: number;
  sides: number;
  back: number;
};

export type DirectionAngleComponent = {
  angle: number;
};

export type OnContactActionComponent = {
  range: number;
  delete_on_touch: boolean;
  action:
    | {
        type: 'gold';
        quantity: number;
      }
    | {
        type: 'stamina';
        quantity: number;
      }
    | {
        type: 'effects';
        effects: Array<{ type: string; duration_s?: number; is_infinite?: boolean }>;
      };
};

// Render components

export type RenderPlayerBlindfoldComponent = {
  blindfold_enabled: boolean;
  visible_tiles_around: number;
};

export type RenderHudComponent = {
  enable_effects: boolean;
  enable_faceset: boolean;
  enable_inventory: boolean;
  enable_status: boolean;
  enable_struggle_bar: boolean;
};

export type RenderSimpleSpriteComponent = {
  is_displayed: boolean;
  sprite: string;
  width: number;
  height: number;
};

export type RenderSimpleSquareComponent = {
  size: number;
  color: string;
};

export type RenderPlayerStruggleBarComponent = {
  enabled: boolean;
  activated: boolean;
  progress_percent: number;
  keys: Array<'up' | 'down' | 'left' | 'right'>;
};

// Sprite controller components

export type SpriteControllerEnemyBasicComponent = {
  is_displayed: boolean;
  sprite_prefix: string;
  direction: 'up' | 'down' | 'left' | 'right';
  fallback_sprite: string;
};

export type SpriteControllerMimicComponent = {
  is_displayed: boolean;
  sprite_prefix: string;
  status: 'idle' | 'attack1' | 'attack2' | 'digest1' | 'digest2';
  fallback_sprite: string;
};

export type SpriteControllerPlayerComponent = {
  is_displayed: boolean;
  sprite_prefix: string;
  type_modifier: string | undefined;
  direction: 'up' | 'down' | 'left' | 'right';
  direction_override: 'ground' | undefined;
  fallback_sprite: string;
};

export type Entity = {
  behavior_ai_component?: BehaviorAIComplexComponent;
  behavior_ai_basic_component?: BehaviorAIBasicComponent;
  behavior_ai_mimic_component?: BehaviorAIMimicComponent;
  behavior_keyboard_component?: BehaviorKeyboardComponent;

  controller_camera_component?: ControllerCameraComponent;
  controller_door_component?: ControllerDoorComponent;
  controller_movement_component?: ControllerMovementComponent;
  controller_stamina_regen_component?: ControllerStaminaRegenComponent;

  angle_component?: DirectionAngleComponent;
  basic_enemy_ai_detection_component?: BasicEnemyAIDetectionComponent;
  collision_component?: CollisionComponent;
  contact_component?: PlayerContactComponent;
  contained_component?: ContainedComponent;
  effects_component?: PlayerEffectsComponent;
  faceset_component?: PlayerFacesetComponent;
  inventory_component?: PlayerInventoryComponent;
  mimic_ai_detection_component?: MimicAIDetectionComponent;
  on_contact_action_component?: OnContactActionComponent;
  position_component?: PositionComponent;
  stamina_component?: StaminaComponent;
  vision_component?: DirectionVisionComponent;

  render_blindfold_component?: RenderPlayerBlindfoldComponent;
  render_hud_component?: RenderHudComponent;
  render_simple_sprite_component?: RenderSimpleSpriteComponent;
  render_simple_square_component?: RenderSimpleSquareComponent;
  render_struggle_bar_component?: RenderPlayerStruggleBarComponent;

  sprite_controller_enemy_basic_component?: SpriteControllerEnemyBasicComponent;
  sprite_controller_mimic_component?: SpriteControllerMimicComponent;
  sprite_controller_player_component?: SpriteControllerPlayerComponent;
};
