import { getDistance } from '../distance.mjs';
import { getNextPathfindingPoint } from '../pathfinding.mjs';

/**
 * @typedef { import("../../engine.mjs").Game } Game
 * @typedef { import("../../types").Entity } Entity
 */

/**
 * Executes the AI Patrol Behavior for the given entity
 * @param {Game} game
 * @param {Entity} entity
 * @param {number} elapsedTicks
 * @returns { string | undefined }
 */
export function patrolAIBehavior(game, entity, elapsedTicks) {
  if (!entity.behavior_ai_component) return;
  if (entity.behavior_ai_component.status.current_state !== 'patrol') return;
  if (!entity.behavior_ai_component.configuration.patrol) return;
  if (!entity.controller_movement_component) return;
  if (!entity.position_component) return;
  if (!entity.angle_component) return;
  const target_patrol_point =
    entity.behavior_ai_component.configuration.patrol.waypoints[
      entity.behavior_ai_component.status.current_patrol_index
    ];
  const distance_to_next_point = getDistance(entity.position_component, target_patrol_point);
  if (distance_to_next_point > entity.behavior_ai_component.configuration.patrol.range_until_point_reached) {
    // Get the next point
    const target_pathfinding_point = getNextPathfindingPoint(game, entity.position_component, target_patrol_point);
    // Move towards the point
    const target_angle = target_pathfinding_point
      ? Math.atan2(
          entity.position_component.y - target_pathfinding_point.y,
          entity.position_component.x - target_pathfinding_point.x
        )
      : 0;
    entity.controller_movement_component.delta_x =
      -entity.behavior_ai_component.configuration.patrol.speed * Math.cos(target_angle) * elapsedTicks;
    entity.controller_movement_component.delta_y =
      -entity.behavior_ai_component.configuration.patrol.speed * Math.sin(target_angle) * elapsedTicks;
    entity.angle_component.angle = target_angle;
    if (entity.angle_component.angle < 0) entity.angle_component.angle += 2 * Math.PI;
    entity.angle_component.angle %= 2 * Math.PI;
    return;
  }

  // Pick the next patrol point
  entity.behavior_ai_component.status.current_patrol_index++;
  entity.behavior_ai_component.status.current_patrol_index %=
    entity.behavior_ai_component.configuration.patrol.waypoints.length;

  entity.controller_movement_component.delta_x = 0;
  entity.controller_movement_component.delta_y = 0;
  return 'patrol_point_reached';
}
