/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").PositionComponent } PositionComponent
 * @typedef { { x: number, y: number, width: number, height: number } } AxisAlignedBoundingBox
 */

/**
 * Returns true if the given bounding box collides with a solid part of the map
 * @param {Game} game The game to check against
 * @param {AxisAlignedBoundingBox} entity_bb The entity bounding box
 * @param {boolean} [visionOnly] Whether to check all collisions or only the vision collisions
 */
export function collidesWithMap(game, entity_bb, visionOnly) {
  const xStartIndex = Math.floor(entity_bb.x / game.mapData.tiles.tile_width);
  const xEndIndex = Math.ceil((entity_bb.x + entity_bb.width) / game.mapData.tiles.tile_width);
  const yStartIndex = Math.floor(entity_bb.y / game.mapData.tiles.tile_height);
  const yEndIndex = Math.ceil((entity_bb.y + entity_bb.height) / game.mapData.tiles.tile_height);
  for (let y = yStartIndex; y < yEndIndex; ++y) {
    const yRow = game.mapData.tiles.content[y];
    if (!yRow) return true;
    for (let x = xStartIndex; x < xEndIndex; ++x) {
      const tile = yRow[x];
      if (!visionOnly && game.mapData.tiles.tiles_prevent_movement[tile]) return true;
      if (visionOnly && game.mapData.tiles.tiles_prevent_vision[tile]) return true;
    }
  }
  return false;
}

/**
 * Returns true if the given bounding box collides with an entity which has collision data
 * @param {Game} game The game to check against
 * @param {AxisAlignedBoundingBox} entity_bb The entity bounding box
 */
export function collidesWithEntity(game, entity_bb) {
  for (const entity of game.entities) {
    if (!entity.collision_component) continue;
    if (!entity.collision_component.is_enabled) continue;
    if (!entity.position_component) continue;
    const startX = entity.position_component.x + entity.collision_component.x_offset;
    const startY = entity.position_component.y + entity.collision_component.y_offset;
    if (
      entity_bb.x < startX + entity.collision_component.width &&
      entity_bb.x + entity_bb.width > startX &&
      entity_bb.y < startY + entity.collision_component.height &&
      entity_bb.y + entity_bb.height > startY
    ) {
      return true;
    }
  }
  return false;
}
