/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").PositionComponent } PositionComponent
 * @typedef { import("../types").CollisionComponent } CollisionComponent
 */

import { getDistance } from './distance.mjs';

/**
 * Returns the next position towards the given target.
 * @param {Game} game The game
 * @param {PositionComponent} origin The current position
 * @param {PositionComponent} target The target position
 * @returns {PositionComponent | undefined} The next target to aim for, or undefined if there is no path
 */
export function getNextPathfindingPoint(game, origin, target) {
  if (getDistance(origin, target) < game.mapData.tiles.tile_width) {
    return target;
  }
  const entity_collision_data = game.entities
    .map((e) => {
      if (!e.position_component) return undefined;
      if (!e.collision_component) return undefined;
      if (!e.collision_component.is_enabled) return undefined;
      return {
        position: e.position_component,
        collision: e.collision_component,
      };
    })
    .filter(
      /** @return {e is Parameters<typeof getTileBasedPathfinding>[3][0] } */
      (e) => !!e
    );
  return getTileBasedPathfinding(game, origin, target, entity_collision_data);
}

/**
 * Returns the next position towards the given target.
 * @param {Game} game The game
 * @param {PositionComponent} origin The current position
 * @param {PositionComponent} target The target position
 * @param {Array<{ position: PositionComponent, collision: CollisionComponent }>} entity_collision_data The entity positions to consider for the collision stuff out
 * @returns {PositionComponent | undefined} The next target to aim for, or undefined if there is no path
 */
function getTileBasedPathfinding(game, origin, target, entity_collision_data) {
  const base_tiles = game.mapData.tiles.content.map((y_row) =>
    y_row.map((c) => game.mapData.tiles.tiles_prevent_movement[c])
  );
  for (const entity_position of entity_collision_data) {
    // TODO: Also handle obstructed adjascent tiles due to the collision box
    const tile_x = Math.floor(entity_position.position.x / game.mapData.tiles.tile_width);
    const tile_y = Math.floor(entity_position.position.y / game.mapData.tiles.tile_height);
    base_tiles[tile_y][tile_x] = true;
  }
  const target_tile = getTargetTilePosition(
    {
      tile_x: Math.floor(origin.x / game.mapData.tiles.tile_width),
      tile_y: Math.floor(origin.y / game.mapData.tiles.tile_height),
    },
    {
      tile_x: Math.floor(target.x / game.mapData.tiles.tile_width),
      tile_y: Math.floor(target.y / game.mapData.tiles.tile_height),
    },
    base_tiles
  );
  if (!target_tile) return undefined;
  return {
    x: target_tile.tile_x * game.mapData.tiles.tile_width + game.mapData.tiles.tile_width / 2,
    y: target_tile.tile_y * game.mapData.tiles.tile_height + game.mapData.tiles.tile_height / 2,
  };
}

/**
 * @param {{ tile_x: number, tile_y: number }} origin The origin, in tiles coordinates
 * @param {{ tile_x: number, tile_y: number }} target The target, in tiles coordinates
 * @param {boolean[][]} tiles tiles[y][x]. True if not allowed, false if allowed.
 * @returns {{ tile_x: number, tile_y: number } | undefined} The flood result
 */
function getTargetTilePosition(origin, target, tiles) {
  const fill_data = fillTiles(origin, target, tiles);
  if (fill_data[origin.tile_y][origin.tile_x] === -1) return undefined;
  if (fill_data[origin.tile_y][origin.tile_x] === 0) return target;

  const top = fill_data[origin.tile_y - 1][origin.tile_x];
  const left = fill_data[origin.tile_y][origin.tile_x - 1];
  const bottom = fill_data[origin.tile_y + 1][origin.tile_x];
  const right = fill_data[origin.tile_y][origin.tile_x + 1];
  const min = Math.min(...[top, left, bottom, right].filter((t) => t >= 0));

  if (top === min && left === min && fill_data[origin.tile_y - 1][origin.tile_x - 1] >= 0) {
    return {
      tile_x: origin.tile_x - 1,
      tile_y: origin.tile_y - 1,
    };
  }
  if (top === min && right === min && fill_data[origin.tile_y - 1][origin.tile_x + 1] >= 0) {
    return {
      tile_x: origin.tile_x + 1,
      tile_y: origin.tile_y - 1,
    };
  }
  if (bottom === min && left === min && fill_data[origin.tile_y + 1][origin.tile_x - 1] >= 0) {
    return {
      tile_x: origin.tile_x - 1,
      tile_y: origin.tile_y + 1,
    };
  }
  if (bottom === min && right === min && fill_data[origin.tile_y + 1][origin.tile_x + 1] >= 0) {
    return {
      tile_x: origin.tile_x + 1,
      tile_y: origin.tile_y + 1,
    };
  }
  if (top === min) {
    return {
      tile_x: origin.tile_x,
      tile_y: origin.tile_y - 1,
    };
  }
  if (left === min) {
    return {
      tile_x: origin.tile_x - 1,
      tile_y: origin.tile_y,
    };
  }
  if (bottom === min) {
    return {
      tile_x: origin.tile_x,
      tile_y: origin.tile_y + 1,
    };
  }
  if (right === min) {
    return {
      tile_x: origin.tile_x + 1,
      tile_y: origin.tile_y,
    };
  }

  return undefined;
}

/**
 * @param {{ tile_x: number, tile_y: number }} origin The origin, in tiles coordinates
 * @param {{ tile_x: number, tile_y: number }} target The target, in tiles coordinates
 * @param {boolean[][]} tiles tiles[y][x]. True if not allowed, false if allowed.
 * @returns {number[][]} The flood result
 */
function fillTiles(origin, target, tiles) {
  const y_width = tiles.length - 1;
  const x_width = tiles[0].length - 1;
  const queue = [{ x: target.tile_x, y: target.tile_y, value: 0 }];
  const fill_values = tiles.map((t) => t.map((_) => -1));
  while (fill_values[origin.tile_y][origin.tile_x] === -1) {
    const next = queue.shift();
    if (!next) break;
    if (tiles[next.y][next.x]) continue;
    fill_values[next.y][next.x] = next.value;
    if (next.x > 0 && fill_values[next.y][next.x - 1] === -1)
      queue.push({ x: next.x - 1, y: next.y, value: next.value + 1 });
    if (next.x < x_width && fill_values[next.y][next.x + 1] === -1)
      queue.push({ x: next.x + 1, y: next.y, value: next.value + 1 });
    if (next.y > 0 && fill_values[next.y - 1][next.x] === -1)
      queue.push({ x: next.x, y: next.y - 1, value: next.value + 1 });
    if (next.y < y_width && fill_values[next.y + 1][next.x] === -1)
      queue.push({ x: next.x, y: next.y + 1, value: next.value + 1 });
  }
  return fill_values;
}
