import { collidesWithMap } from './collision.mjs';

/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").PositionComponent } PositionComponent
 */

/**
 * Returns the vision box of a given entity.
 * @param {Game} game The game
 * @param {PositionComponent} entity The entity whose vision box to check
 * @param {number} view_angle The current angle of the entity, in radians (0 = left, PI / 2 = bottom, PI = right, etc...)
 * @param {number} distance_in_front The base view distance to use in front of the entity
 * @param {number} distance_around The base view square to use around the enemy
 * @param {number} distance_in_back The base view square to use behind the entity
 * @returns {{
 *   x: number,
 *   y: number ,
 *   width: number,
 *   height: number,
 * }} The entity's vision box, in raw coordinates, not hitting the map.
 */
export function getVisionBox(game, entity, view_angle, distance_in_front, distance_around, distance_in_back) {
  if (view_angle <= Math.PI * (1 / 4) || view_angle >= Math.PI * (1 / 2 + 1 / 2 + 1 / 2 + 1 / 4)) {
    let view_distance = distance_in_front;
    while (
      collidesWithMap(
        game,
        {
          x: entity.x - view_distance + distance_in_back / 2,
          y: entity.y - distance_around / 2,
          width: view_distance,
          height: distance_around,
        },
        true
      )
    ) {
      view_distance -= 10;
      if (view_distance == 0) break;
    }
    return {
      x: entity.x - view_distance + distance_in_back / 2,
      y: entity.y - distance_around / 2,
      width: view_distance,
      height: distance_around,
    };
  }
  if (view_angle > Math.PI * (1 / 2 + 1 / 4) && view_angle <= Math.PI * (1 / 2 + 1 / 2 + 1 / 4)) {
    let view_distance = distance_in_front;
    while (
      collidesWithMap(
        game,
        {
          x: entity.x - distance_in_back / 2,
          y: entity.y - distance_around / 2,
          width: view_distance,
          height: distance_around,
        },
        true
      )
    ) {
      view_distance -= 10;
      if (view_distance == 0) break;
    }
    return {
      x: entity.x - distance_in_back / 2,
      y: entity.y - distance_around / 2,
      width: view_distance,
      height: distance_around,
    };
  }
  if (view_angle > Math.PI * (1 / 4) && view_angle <= Math.PI * (1 / 2 + 1 / 4)) {
    let view_distance = distance_in_front;
    while (
      collidesWithMap(
        game,
        {
          x: entity.x - distance_around / 2,
          y: entity.y - view_distance + distance_in_back / 2,
          width: distance_around,
          height: view_distance,
        },
        true
      )
    ) {
      view_distance -= 10;
      if (view_distance == 0) break;
    }
    return {
      x: entity.x - distance_around / 2,
      y: entity.y - view_distance + distance_in_back / 2,
      width: distance_around,
      height: view_distance,
    };
  }
  if (view_angle > Math.PI * (1 / 2 + 1 / 2 + 1 / 4) && view_angle <= Math.PI * (1 / 2 + 1 / 2 + 1 / 2 + 1 / 4)) {
    let view_distance = distance_in_front;
    while (
      collidesWithMap(
        game,
        {
          x: entity.x - distance_around / 2,
          y: entity.y - distance_in_back / 2,
          width: distance_around,
          height: view_distance,
        },
        true
      )
    ) {
      view_distance -= 10;
      if (view_distance == 0) break;
    }
    return {
      x: entity.x - distance_around / 2,
      y: entity.y - distance_in_back / 2,
      width: distance_around,
      height: view_distance,
    };
  }

  // This should not happen
  throw new Error(`View angle ${view_angle} is not a valid angle value`);
}
